% Prolog part based on clever N-Queens program by T. Fruhwrith

queens(N,Qs):- gen_list(N,Qs), place_queens(N,Qs,_,_).

gen_list(0,[]).
gen_list(N,[_|L]):-
        N>0, M is N-1,
        gen_list(M,L).

place_queens(0,_,_,_).
place_queens(I,Cs,Us,[_|Ds]):-
        I>0, J is I-1,
        place_queens(J,Cs,[_|Us],Ds),
        place_queen(I,Cs,Us,Ds).

place_queen(I,[I|_],[I|_],[I|_]).
place_queen(I,[_|Cs],[_|Us],[_|Ds]):-
        place_queen(I,Cs,Us,Ds).

% Tcl/Tk interactive visualisation

qs(N):-
	queens(N,L),
	call_tcl(display_queens(L)),   % sends a display request
	write('type <p done> when finished <p more> otherwise'),nl,
	in(call_prolog(done)),         % waits for `done' or `more'
	!.

/****************************************************************

This program illustrates two simple-minded but useful features:
1. Prolog lists are sent to Tcl as Tcl-lists
2. Prolog simply pumps solutions one by one using their most natural
   representation through "call_tcl(display_queens(L))".
   Synchronization is done as follows:

   - in/1 blocks the prolog process until Tcl confirms it want another
     one

Messages have special `headers' like "call_tcl" and "call_prolog" to
avoid spurious intraction due to errors on any side.

Note that you can also control the program from Tcl's command line
by typing in "p more" or "p done" messages

*******************************************************************/
