#!/usr/local/bin/wish -f

# Tcl/Tk side of the BinProlog-Tcl/Tk interface
# tested on tcl7.3 with tk3.6 patched with addinput-3.6a

# USAGE: under the wish tcl/tk shell do:
#
# % start_prolog
# % q {PrologQuery}
# % p {QuietPrologGoal}
# % halt_prolog
#

global f fpid

# tests what Tcl would send to Prolog
proc start_test {} {
  global f
  set f stdout
}

# terminates the Prolog process
proc halt_prolog {} {
  global f fpid
  catch {removeinput $f}
  catch {close $f}
  catch {exec sh -c "kill -9 $fpid"}
  catch {puts "Removed input on $f from process $fpid"}
}

# connects to a new Prolog process
proc start_prolog {} {
  halt_prolog
  global f fpid
  set f [open "|ru -q5 server.bp" r+]
  set fpid [pid $f]
  addinput $f "bpInOut %% %E %F"
}

# writes "Something." to Prolog
proc write_prolog {s} {
  global f
  puts $f $s.
  flush $f
}

# sends to prolog a goal to be quietly evaluated
proc call_prolog {l} {
  write_prolog "call_prolog(($l))"
}

# sends to prolog a query with variables to be answered
proc query_prolog {l} {
  write_prolog "query_prolog(($l))"
}

# shorthand for call_prolog
proc p {l} {
  call_prolog $l
}

# shorthand for query_prolog
proc q {l} {
  query_prolog $l
}

# true if there's a tcl call coming from Prolog
proc prolog_is_calling_tcl {l} {
  return [expr 0 == [string first "call_tcl " [string trimleft $l]]]
}

# evaluates a call to tcl from Prolog
proc call_tcl {l} {
  return [eval $l]
}

# detects if an input line contains a tcl call 
# (i.e. when it starts with "call_tcl ")
# if yes then it executes the line
# otherwise it prints it out
proc eval_line {l} {
if {[prolog_is_calling_tcl $l]} {
    call_tcl $l
  } else {
    puts $l
  }
}

# spies (in background) on input coming from Prolog
# and passes it to eval_line
# the prolog process should flush the output when doing "nl/0" 

proc bpInOut {token events fileid} {
    if {"$events" != "READ"} {
        error "Unexpected events: $events"
    } elseif {[gets $fileid line] < 0} {
        halt_prolog
    } else {
       eval_line $line
    }
}
