% -----------------------------------------------------------------
% LIBRARY of basic predicates

op(Pri,Assoc,Name):-op0(Name,Assoc,Pri).

current_op(Pri,Assoc,Name):-nonvar(Name),nonvar(Assoc),!,
  val(Name,Assoc,Pri).
current_op(Pri,Assoc,Name):-
  for(I,0,10000),
  (get_op(I,N,A,P)->N=Name,A=Assoc,P=Pri;!,fail).

get_op(I,Name,Assoc,Pri):- 
	val(op_name,I,Name),
	val(op_assoc,I,Assoc),
	val(op_pri,I,Pri).

is_interactive:-current_op(1200,fx,('?-')).

interactive(yes):-op(1200,fx,('?-')).
interactive(no):-op(1199,fx,('?-')).

expand_term(C,E):-portable_expand_term(C,E).
std_expand_term(C,D):-portable_expand_term(C,D).

metacall(G):-call_ifdef(meta_interpreter(G),G).

% true.
% call(X):-X.

call_ifdef(G,_):-is_compiled(G),!,G.
call_ifdef(_,Else):-Else.

X=X.

A->B :- A,!,B.

A->B ; C :- !,if(A,B,C).
X ; _ :-X.
_ ; Y :-Y.

or(A,_):-A.
or(_,B):-B.

if(A,B,_):-A,!,B.
if(_,_,C):-C.

(X,Y):-X,Y.

\+(X):-X,!,fail.
\+(_).

halt:-halt(0).

repeat.
repeat:-repeat.

findall_workhorse(X,G,_):-
	lift_heap,
	G,
	findall_store_heap(X).
findall_workhorse(_,_,Xs):-
	findall_load_heap(Xs).	

findall(X,G,Xs,End):-findall_workhorse(X,G,[End|Xs]).

findall(X,G,Xs):-findall(X,G,Xs,[]).

findall(G,Gs):-findall(G,G,Gs).

gc_read(R):-findall(X,read(X),[R]).

gc_call(G):-findall(G,G,Gs),member(G,Gs).

for(Min,Min,Max):-Min=<Max.
for(I,Min,Max):-
        Min<Max,
        Min1 is Min+1,
        for(I,Min1,Max).

numbervars('$VAR'(N0), N0, N) :- !, N is N0+1.
numbervars(X, N0, N) :- atomic(X), !, N0=N.
numbervars([X|Xs], N0, N) :- !,
        numbervars(X, N0, N1),
        numbervars(Xs, N1, N).
numbervars(X, N0, N) :-
        functor(X, _, A),
        numbervars(0, A, X, N0, N).

numbervars(A, A, _, N0, N) :- !, N0=N.
numbervars(A0, A, X, N0, N) :-
        A1 is A0+1,
        arg(A1, X, X1),
        numbervars(X1, N0, N1),
        numbervars(A1, A, X, N1, N).

ground(T):-numbervars(T,0,N),N>0,!,fail.
ground(_).

copy_term(T,CT):-copy_term(0,T,CT).
save_term(T,CT):-copy_term(1,T,CT).

seen:-seen_told(0).
told:-seen_told(1).
flush:-seen_told(2).

see_or_fail(File):-see_tell(0,File).

see(File):-see_tell(0,File),!.
see(File):-user_error(unable_to_see,File).

seeing(F):-seeing_telling(0,F).
telling(F):-seeing_telling(1,F).

tell(File):-see_tell(1,File),!.
tell(File):-user_error(unable_to_tell,File).

sread(I,O):-string_op(0,I,O).
swrite(I,O):-string_op(1,I,O).

system([]):-!.
system([X|Xs]):-!,name(Command,[X|Xs]),shell(Command).
system(Command):-atomic(Command),shell(Command).

[File]:-compile(File).

compile(File):-
	find_file(File,F),
	compile_mem(F),fail.

find_file(File,_):-compound(File),!,errmes('bad file name',File).
find_file(File,NewFile):-
	seeing(CF),
	find_file1(CF,File,NewFile).

find_file1(CF,File,NewFile):-
	see_a_file(
		["","progs/","myprogs/"],File,
		[".pl",".pro",""],NewFile),
	!,
	see(CF).
find_file1(CF,File,_):-
	see(CF),
	errmes('file * *.pl *.pro not found in . progs myprogs:',File).

see_a_file(Prefs,File,Sufs,NewFile):-
	member(Suf,Sufs),
	member(Pref,Prefs),
	name(File,L),
	det_append(L,Suf,R),
	det_append(Pref,R,Fname),
	name(NewFile,Fname),
	see_or_fail(NewFile),
	seen.

statistics(Area,[Used,Free]):-stat_dict(Area,No),stat0(Used,Free,No).

statistics:-
	statistics(Name,Data),
	cwrite(Name),cwrite(=),cwrite(Data),nl,
	fail.
statistics.	

atom(X):-integer(X),!,fail.
atom(X):-atomic(X).

float(X):-nonvar(X),X='$float'(_,_,_).

number(X):-integer(X),!.
number(X):-float(X).

compound(X):-nonvar(X), \+(atomic(X)).

=..(T,[F|Xs]):-nonvar(T),!,functor(T,F,N),term2list(1,N,T,Xs).
=..(T,[F|Xs]):-get_length(Xs,0,N),!,functor(T,F,N),term2list(1,N,T,Xs).

term2list(I,N,_,R):-I>N,!,R=[].
term2list(I,N,T,[X|Xs]):-I1 is I+1,arg(I,T,X),term2list(I1,N,T,Xs).

length(L,N):-var(N),!,get_length(L,0,N).
length(L,N):-make_length(L,0,N).

get_length([],I,I).
get_length([_|L],I0,I):-I1 is I0+1,get_length(L,I1,I).

make_length([],I,I):-!.
make_length([_|L],I0,I):-I0<I,I1 is I0+1,make_length(L,I1,I).

tab(Expr) :- expr(Expr,N),integer(N),!,tab0(N).
tab(Bad) :- errmes('in tab/1: should evaluate to an integer', Bad).

tab0(N):-for(_,1,N),put(32),fail.
tab0(_).

get(R):-repeat,get0(X),(X>32;X<0),!,R=X.

% !! use something like Y=3+4, X is 1+expr(Y) instead of
% !! Y=3+4, X is 1+Y which will not work in compiled code

X is E:-expr(E,R),!,X=R.
X is E:-user_error(error_in_is,X is E).

expr(E,R):-atomic(E),!,R=E.
expr(E,R):-float(E),!,R=E.
expr(E,_Error):-var(E),!,errmes(variable_in_is,E).
expr(E,R):-functor(E,Op,2),!,
	arg(1,E,E1),arg(2,E,E2),
	expr(E1,X1),
	expr(E2,X2),
	G=..[Op,X1,X2,R],
	G,!.
expr(E,R):-functor(E,Op,1),
	arg(1,E,E1),
	expr(E1,X1),
	G=..[Op,X1,R],
	G,!.

compare(R,X,Y):-compare0(X,Y,R).

+(X,X).
-(X,N):- -(0,X,N).
\(X,N):- \(0,X,N).

A==B :- compare0(A,B,=).
A\==B :- compare0(A,B,R),'$noteq'(R).

A @< B :- compare0(A,B,<).
A @> B :- compare0(A,B,>).
A @=< B :- compare0(A,B,R),'$lesseq'(R).
A @>= B :- compare0(A,B,R),'$gteq'(R).

A < B :-expr(A,X), expr(B,Y), less(X,Y).

A > B :-expr(A,X), expr(B,Y), greater(X,Y).

A =< B :-expr(A,X), expr(B,Y), less_eq(X,Y).

A >= B :-expr(A,X), expr(B,Y), greater_eq(X,Y).

A =:= B :-expr(A,X), expr(B,Y), arith_eq(X,Y).

A =\= B :-expr(A,X), expr(B,Y), arith_dif(X,Y).

'$lesseq'(<).
'$lesseq'(=).

'$gteq'(>).
'$gteq'(=).

'$noteq'(<).
'$noteq'(>).

% BinProlog USER LEVEL naming primitives
% DO NOT rely on def/3, set/3, val/3,rm/2. Thay are unsafe
% and their implementation may change in the future...

/********* BASIC operations: safe with respect of bb_gc ***********/

% WARNING: may be moved to C, do not use their componnents directly

bb_def(A,B,X):-saved(X,S),def(A,B,S).
bb_set(A,B,X):-saved(X,S),set(A,B,S).
bb_val(A,B,CX):-val(A,B,X),copy_term(X,CX).
bb_rm(A,B):-rm(A,B).

/******************DERIVED OPERATIONS ***********************/

bb_let(A,B,X):-val(A,B,_),!,bb_set(A,B,X).
bb_let(A,B,X):-bb_def(A,B,X).

bb_def(A,X):-bb_def(A,A,X).
bb_set(A,X):-bb_set(A,A,X).
bb_val(A,X):-bb_val(A,A,X).
bb_rm(A):-bb_rm(A,A).
bb_let(A,X):-bb_let(A,A,X).

saved(X,S):-save_term(X,NewX),!,S=NewX.
saved(X,S):-bb_gc,save_term(X,NewX),!,S=NewX.
saved(_,_):-bb_fail(saved/2).
saved(_,_):-
	restart,
	abort.

% GLOBAL LOGICAL VARIABLES

'=>'(K1#K2,X):-!,lval(K1,K2,X).
'=>'(K,X):-lval(K,K,X).

/* bboard garbage collector */

bb_gc:-
	bb_list(1,B),
	copy_term(B,NewB),
	bb_reset,
	bb_put_back(NewB).

bb_put_back([]).
bb_put_back([P,N,F,K,V|Xs]):-
        functor(O,P,N),
        functor(M,F,K),
	bb_put_1(O,M,V),
	bb_put_back(Xs).

bb_put_1(O,M,V):-save_term(V,NewV),set(O,M,NewV),!,NewV=V.
bb_put_1(O,M,_):-bb_fail(O+M).

bb_fail(X):-
	statistics(bboard,[_,Z]),
	user_error('blackboard overflow, left only'(Z),culprit(X)).

is_dynamic(P):-val(P,'$first',_).

% An easy patch to read.pl for basic float input - Paul Tarau, Oct. 1993

try_float(Ds,[atom('.'),integer(box(Fraq,FDs))|Ts],I,[atom(Float)|Ts1]):-!,
	try_float_exp(Ts,Ts1,Exp),name(NFraq,FDs),
	input_float(I,NFraq,Exp,Float).
try_float(Ds,Tokens,I,[integer(box(I,Ds))|Tokens]).

try_float_exp([atom(e)|Ts],Ts1,Exp):-try_float_exp1(Ts,Ts1,Exp),!.
try_float_exp([atom(EN)|Ts],Ts,Exp):-
	name(EN,[0'e|NL]),name(Exp,NL),integer(Exp),!.
try_float_exp(Ts,Ts,0).

try_float_exp1([atom('-'),integer(box(Exp,_))|Ts],Ts,Exp1):-!,
	Exp1 is 0-Exp.
try_float_exp1([atom('+'),integer(box(Exp,_))|Ts],Ts,Exp).

float_minus(F,NegF):-NegF is 0.0-F.

pow(A,X,R):-float_fun2(pow,A,X,R).
log(A,X,R):-float_fun2(log,A,X,R).

% F(X+Y) --> R used for Y=0
exp(X,R):-float_fun(exp,X,R).
log(X,R):-float_fun(log,X,R).
sin(X,R):-float_fun(sin,X,R). 
cos(X,R):-float_fun(cos,X,R).
tan(X,R):-float_fun(tan,X,R).

% F(X)+Y --> R used for Y=0
atan(X,R):-float_fun('T',X,R).
asin(X,R):-float_fun('S',X,R).
acos(X,R):-float_fun('C',X,R).

integer(X,R):-float_fun(integer,X,R).
float(X,R):-R is 0.0+X.

% for portability...
ctime(T):-statistics(runtime,[T,_]).
